from __future__ import annotations

import logging
import typing as t

from globus_sdk import exc, utils
from globus_sdk._types import ScopeCollectionType, UUIDLike
from globus_sdk.authorizers import BasicAuthorizer
from globus_sdk.response import GlobusHTTPResponse

from .._common import stringify_requested_scopes
from ..flow_managers import GlobusAuthorizationCodeFlowManager
from ..response import (
    GetIdentitiesResponse,
    OAuthDependentTokenResponse,
    OAuthTokenResponse,
)
from .base_login_client import AuthLoginClient

log = logging.getLogger(__name__)


class ConfidentialAppAuthClient(AuthLoginClient):
    """
    This is a specialized type of ``AuthLoginClient`` used to represent an App with
    a Client ID and Client Secret wishing to communicate with Globus Auth.
    It must be given a Client ID and a Client Secret, and furthermore, these
    will be used to establish a :class:`BasicAuthorizer <globus_sdk.BasicAuthorizer>`
    for authorization purposes.
    Additionally, the Client ID is stored for use in various calls.

    Confidential Applications are those which have their own credentials for
    authenticating against Globus Auth.

    :param client_id: The ID of the application provided by registration with
        Globus Auth.
    :param client_secret: The secret string to use for authentication. Secrets can be
        generated via the Globus developers interface.

    All other initialization parameters are passed through to ``BaseClient``.

    .. automethodlist:: globus_sdk.ConfidentialAppAuthClient
    """

    def __init__(
        self,
        client_id: UUIDLike,
        client_secret: str,
        environment: str | None = None,
        base_url: str | None = None,
        app_name: str | None = None,
        transport_params: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            authorizer=BasicAuthorizer(str(client_id), client_secret),
            environment=environment,
            base_url=base_url,
            app_name=app_name,
            transport_params=transport_params,
        )

    def get_identities(
        self,
        *,
        usernames: t.Iterable[str] | str | None = None,
        ids: t.Iterable[UUIDLike] | UUIDLike | None = None,
        provision: bool = False,
        query_params: dict[str, t.Any] | None = None,
    ) -> GetIdentitiesResponse:
        """
        Perform a call to the Get Identities API using the direct client
        credentials of this client.

        This method is considered deprecated -- callers should instead use client
        credentials to get a token and then use that token to call the API via a
        :class:`~.AuthClient`.

        :param usernames: A username or list of usernames to lookup. Mutually exclusive
            with ``ids``
        :param ids: An identity ID or list of IDs to lookup. Mutually exclusive
            with ``usernames``
        :param provision: Create identities if they do not exist, allowing clients to
            get username-to-identity mappings prior to the identity being used
        :param query_params: Any additional parameters to be passed through
            as query params.
        """
        exc.warn_deprecated(
            "ConfidentialAuthClient.get_identities() is deprecated. "
            "Get a token via `oauth2_client_credentials_tokens` "
            "and use that to call the API instead."
        )

        if query_params is None:
            query_params = {}

        if usernames is not None:
            query_params["usernames"] = utils.commajoin(usernames)
            query_params["provision"] = (
                "false" if str(provision).lower() == "false" else "true"
            )
        if ids is not None:
            query_params["ids"] = utils.commajoin(ids)

        return GetIdentitiesResponse(
            self.get("/v2/api/identities", query_params=query_params)
        )

    def oauth2_client_credentials_tokens(
        self,
        requested_scopes: ScopeCollectionType | None = None,
    ) -> OAuthTokenResponse:
        r"""
        Perform an OAuth2 Client Credentials Grant to get access tokens which
        directly represent your client and allow it to act on its own
        (independent of any user authorization).
        This method does not use a ``GlobusOAuthFlowManager`` because it is not
        at all necessary to do so.

        :param requested_scopes: The scopes on the token(s) being requested. Defaults to
            ``openid profile email urn:globus:auth:scope:transfer.api.globus.org:all``

        For example, with a Client ID of "CID1001" and a Client Secret of
        "RAND2002", you could use this grant type like so:

        >>> client = ConfidentialAppAuthClient("CID1001", "RAND2002")
        >>> tokens = client.oauth2_client_credentials_tokens()
        >>> transfer_token_info = (
        ...     tokens.by_resource_server["transfer.api.globus.org"])
        >>> transfer_token = transfer_token_info["access_token"]
        """
        log.info("Fetching token(s) using client credentials")
        requested_scopes_string = stringify_requested_scopes(requested_scopes)
        return self.oauth2_token(
            {"grant_type": "client_credentials", "scope": requested_scopes_string}
        )

    def oauth2_start_flow(
        self,
        redirect_uri: str,
        requested_scopes: ScopeCollectionType | None = None,
        *,
        state: str = "_default",
        refresh_tokens: bool = False,
    ) -> GlobusAuthorizationCodeFlowManager:
        """
        Starts or resumes an Authorization Code OAuth2 flow.

        Under the hood, this is done by instantiating a
        :class:`GlobusAuthorizationCodeFlowManager
        <.GlobusAuthorizationCodeFlowManager>`

        :param redirect_uri: The page that users should be directed to after
            authenticating at the authorize URL.
        :param requested_scopes: The scopes on the token(s) being requested. Defaults to
            ``openid profile email urn:globus:auth:scope:transfer.api.globus.org:all``
        :param state: This string allows an application to pass information back to
            itself in the course of the OAuth flow. Because the user will navigate away
            from the application to complete the flow, this parameter lets the app pass
            an arbitrary string from the starting page to the ``redirect_uri``
        :param refresh_tokens: When True, request refresh tokens in addition to access
            tokens. [Default: ``False``]

        .. tab-set::

            .. tab-item:: Example Usage

                You can see an example of this flow :ref:`in the usage examples
                <examples_three_legged_oauth_login>`.

            .. tab-item:: API Info

                The Authorization Code Grant flow is described
                `in the Globus Auth Specification
                <https://docs.globus.org/api/auth/developer-guide/#obtaining-authorization>`_.
        """
        log.info("Starting OAuth2 Authorization Code Grant Flow")
        self.current_oauth2_flow_manager = GlobusAuthorizationCodeFlowManager(
            self,
            redirect_uri,
            requested_scopes=requested_scopes,
            state=state,
            refresh_tokens=refresh_tokens,
        )
        return self.current_oauth2_flow_manager

    def oauth2_get_dependent_tokens(
        self,
        token: str,
        *,
        refresh_tokens: bool = False,
        additional_params: dict[str, t.Any] | None = None,
    ) -> OAuthDependentTokenResponse:
        """
        Does a `Dependent Token Grant
        <https://docs.globus.org/api/auth/reference/#dependent_token_grant_post_v2_oauth2_token>`_
        against Globus Auth.
        This exchanges a token given to this client for a new set of tokens
        which give it access to resource servers on which it depends.
        This grant type is intended for use by Resource Servers playing out the
        following scenario:

          1. User has tokens for Service A, but Service A requires access to
             Service B on behalf of the user
          2. Service B should not see tokens scoped for Service A
          3. Service A therefore requests tokens scoped only for Service B,
             based on tokens which were originally scoped for Service A...

        In order to do this exchange, the tokens for Service A must have scopes
        which depend on scopes for Service B (the services' scopes must encode
        their relationship). As long as that is the case, Service A can use
        this Grant to get those "Dependent" or "Downstream" tokens for Service B.

        :param token: A Globus Access Token as a string
        :param refresh_tokens: When True, request dependent refresh tokens in addition
            to access tokens. [Default: ``False``]
        :param additional_params: Additional parameters to include in the request body
        """
        log.info("Getting dependent tokens from access token")
        log.debug(f"additional_params={additional_params}")
        form_data = {
            "grant_type": "urn:globus:auth:grant_type:dependent_token",
            "token": token,
        }
        # the internal parameter is 'access_type', but using the name 'refresh_tokens'
        # is consistent with the rest of the SDK and better communicates expectations
        # back to the user than the OAuth2 spec wording
        if refresh_tokens:
            form_data["access_type"] = "offline"
        if additional_params:
            form_data.update(additional_params)

        return self.oauth2_token(form_data, response_class=OAuthDependentTokenResponse)

    def oauth2_token_introspect(
        self, token: str, *, include: str | None = None
    ) -> GlobusHTTPResponse:
        """
        Get information about a Globus Auth token.

        :param token: An Access Token as a raw string, being evaluated
        :param include: A value for the ``include`` parameter in the request body.
            Default is to omit the parameter.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    ac = globus_sdk.ConfidentialAppAuthClient(CLIENT_ID, CLIENT_SECRET)
                    ac.oauth2_token_introspect("<token_string>")

                Get information about a Globus Auth token including the full identity
                set of the user to whom it belongs

                .. code-block:: python

                    ac = globus_sdk.ConfidentialAppAuthClient(CLIENT_ID, CLIENT_SECRET)
                    data = ac.oauth2_token_introspect("<token_string>", include="identity_set")
                    for identity in data["identity_set"]:
                        print('token authenticates for "{}"'.format(identity))

            .. tab-item:: API Info

                ``POST /v2/oauth2/token/introspect``

                .. extdoclink:: Token Introspection
                    :ref: auth/reference/#token_introspection_post_v2_oauth2_token_introspect
        """  # noqa: E501
        log.info("Checking token validity (introspect)")
        body = {"token": token}
        if include is not None:
            body["include"] = include
        return self.post("/v2/oauth2/token/introspect", data=body, encoding="form")

    def create_child_client(
        self,
        name: str,
        *,
        public_client: bool | utils.MissingType = utils.MISSING,
        client_type: (
            t.Literal[
                "client_identity",
                "confidential_client",
                "globus_connect_server",
                "public_installed_client",
                "hybrid_confidential_client_resource_server",
                "resource_server",
            ]
            | utils.MissingType
        ) = utils.MISSING,
        visibility: t.Literal["public", "private"] | utils.MissingType = utils.MISSING,
        redirect_uris: t.Iterable[str] | utils.MissingType = utils.MISSING,
        terms_and_conditions: str | utils.MissingType = utils.MISSING,
        privacy_policy: str | utils.MissingType = utils.MISSING,
        required_idp: UUIDLike | utils.MissingType = utils.MISSING,
        preselect_idp: UUIDLike | utils.MissingType = utils.MISSING,
        additional_fields: dict[str, t.Any] | utils.MissingType = utils.MISSING,
    ) -> GlobusHTTPResponse:
        """
        Create a new client. Requires the ``manage_projects`` scope.

        :param name: The display name shown to users on consents. May not contain
            linebreaks.
        :param public_client: This is used to infer which OAuth grant_types the client
            will be able to use. Should be false if the client is capable of keeping
            secret credentials (such as clients running on a server) and true if it is
            not (such as native apps). After creation this value is immutable. This
            option is mutually exclusive with ``client_type``, exactly one must be
            given.
        :param client_type: Defines the type of client that will be created. This
            option is mutually exclusive with ``public_client``, exactly one must
            be given.

            .. dropdown:: Values for ``client_type``

                .. list-table::

                    * - ``"confidential_client"``
                      - Applications that are OAuth confidential clients, and can
                        manage a client secret and requests for user consent.
                    * - ``"public_installed_client"``
                      - Applications that are OAuth public clients or native
                        applications that are distributed to users, and thus cannot
                        manage a client secret.
                    * - ``"client_identity"``
                      - Applications that authenticate and act as the application
                        itself. These applications are used for automation and as
                        service or community accounts, and do NOT act on behalf of
                        other users. Also known as a "Service Account".
                    * - ``"resource_server"``
                      - An API (OAuth resource server) that uses Globus Auth tokens for
                        authentication. Users accessing the service login via Globus and
                        consent for the client to use your API.
                    * - ``"globus_connect_server"``
                      - Create a client that will service as a Globus Connect Server
                        endpoint.
                    * - ``"hybrid_confidential_client_resource_server"``
                      - A client which can use any behavior with Globus Auth - an
                        application (confidential or public client), service account,
                        or API.

        :param visibility: If set to "public", any authenticated entity can view it.
            When set to "private", only entities in the same project as the client can
            view it.
        :param redirect_uris: list of URIs that may be used in OAuth authorization
            flows.
        :param terms_and_conditions: URL of client's terms and conditions.
        :param privacy_policy: URL of client's privacy policy.
        :param required_idp: In order to use this client a user must have an identity
            from this IdP in their identity set.
        :param preselect_idp: This pre-selects the given IdP on the Globus Auth login
            page if the user is not already authenticated.
        :param additional_fields: Any additional parameters to be passed through.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> project_id = ...
                    >>> r = ac.create_child_client(
                    ...     "My client",
                    ...     True,
                    ...     True,
                    ... )
                    >>> client_id = r["client"]["id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_child_client

            .. tab-item:: API Info

                ``POST /v2/api/clients``

                .. extdoclink:: Create Client
                    :ref: auth/reference/#create_client
        """
        # Must specify exactly one of public_client or client_type
        if public_client is not utils.MISSING and client_type is not utils.MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.create_client does not take both "
                "'public_client' and 'client_type'. These are mutually exclusive."
            )
        if public_client is utils.MISSING and client_type is utils.MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.create_client requires either 'public_client' or "
                "'client_type'."
            )

        body: dict[str, t.Any] = {
            "name": name,
            "visibility": visibility,
            "required_idp": required_idp,
            "preselect_idp": preselect_idp,
            "public_client": public_client,
            "client_type": client_type,
        }
        if not isinstance(redirect_uris, utils.MissingType):
            body["redirect_uris"] = list(utils.safe_strseq_iter(redirect_uris))

        # terms_and_conditions and privacy_policy must both be set or unset
        if bool(terms_and_conditions) ^ bool(privacy_policy):
            raise exc.GlobusSDKUsageError(
                "terms_and_conditions and privacy_policy must both be set or unset"
            )
        links: dict[str, str | utils.MissingType] = {
            "terms_and_conditions": terms_and_conditions,
            "privacy_policy": privacy_policy,
        }
        if terms_and_conditions or privacy_policy:
            body["links"] = links

        if not isinstance(additional_fields, utils.MissingType):
            body.update(additional_fields)

        return self.post("/v2/api/clients", data={"client": body})



from __future__ import annotations

import typing as t

from . import _serializable, _validators


class GlobusAuthorizationParameters(_serializable.Serializable):
    """
    Represents authorization parameters that can be used to instruct a client
    which additional authorizations are needed in order to complete a request.

    :ivar session_message: A message to be displayed to the user.
    :vartype session_message: str, optional

    :ivar session_required_identities: A list of identities required for the
        session.
    :vartype session_required_identities: list of str, optional

    :ivar session_required_policies: A list of policies required for the
        session.
    :vartype session_required_policies: list of str, optional

    :ivar session_required_single_domain: A list of domains required for the
        session.
    :vartype session_required_single_domain: list of str, optional

    :ivar session_required_mfa: Whether MFA is required for the session.
    :vartype session_required_mfa: bool, optional

    :ivar required_scopes: A list of scopes for which consent is required.
    :vartype required_scopes: list of str, optional

    :ivar prompt: The OIDC 'prompt' parameter, for which Globus Auth currently supports
        the values 'login' and 'none'.
    :vartype prompt: str, optional

    :ivar extra: A dictionary of additional fields that were provided. May
        be used for forward/backward compatibility.
    :vartype extra: dict
    """

    def __init__(
        self,
        *,
        session_message: str | None = None,
        session_required_identities: list[str] | None = None,
        session_required_policies: list[str] | None = None,
        session_required_single_domain: list[str] | None = None,
        session_required_mfa: bool | None = None,
        required_scopes: list[str] | None = None,
        prompt: str | None = None,
        extra: dict[str, t.Any] | None = None,
    ):
        self.session_message = _validators.opt_str("session_message", session_message)
        self.session_required_identities = _validators.opt_str_list(
            "session_required_identities", session_required_identities
        )
        self.session_required_policies = _validators.opt_str_list(
            "session_required_policies", session_required_policies
        )
        self.session_required_single_domain = _validators.opt_str_list(
            "session_required_single_domain", session_required_single_domain
        )
        self.session_required_mfa = _validators.opt_bool(
            "session_required_mfa", session_required_mfa
        )
        self.required_scopes = _validators.opt_str_list(
            "required_scopes", required_scopes
        )
        self.prompt = _validators.opt_str("prompt", prompt)
        self.extra = extra or {}

        # Enforce that the error contains at least one of the fields we expect
        requires_at_least_one = [
            name for name in self._supported_fields() if name != "session_message"
        ]
        if all(
            getattr(self, field_name) is None for field_name in requires_at_least_one
        ):
            raise _validators.ValidationError(
                "Must include at least one supported authorization parameter: "
                + ", ".join(requires_at_least_one)
            )


class GlobusAuthRequirementsError(_serializable.Serializable):
    """
    Represents a Globus Auth Requirements Error.

    A Globus Auth Requirements Error is a class of error that is returned by Globus
    services to indicate that additional authorization is required in order to complete
    a request and contains information that can be used to request the appropriate
    authorization.

    :ivar code: The error code for this error.
    :vartype code: str

    :ivar authorization_parameters: The authorization parameters for this error.
    :vartype authorization_parameters: GlobusAuthorizationParameters

    :ivar extra: A dictionary of additional fields that were provided. May
        be used for forward/backward compatibility.
    :vartype extra: dict
    """

    def __init__(
        self,
        code: str,
        authorization_parameters: dict[str, t.Any] | GlobusAuthorizationParameters,
        *,
        extra: dict[str, t.Any] | None = None,
    ):
        self.code = _validators.str_("code", code)
        self.authorization_parameters = _validators.instance_or_dict(
            "authorization_parameters",
            authorization_parameters,
            GlobusAuthorizationParameters,
        )
        self.extra = extra or {}



from globus_sdk import response


class IterableFlowsResponse(response.IterableResponse):
    """
    An iterable response containing a "flows" array.
    """

    default_iter_key = "flows"


class IterableRunsResponse(response.IterableResponse):
    """
    An iterable response containing a "runs" array.
    """

    default_iter_key = "runs"


class IterableRunLogsResponse(response.IterableResponse):
    """
    An iterable response containing an "entries" array.
    """

    default_iter_key = "entries"



from __future__ import annotations

import typing as t

from globus_sdk import _guards

from ._serializable import Serializable

S = t.TypeVar("S", bound=Serializable)


class ValidationError(ValueError):
    pass


def str_(name: str, value: t.Any) -> str:
    if isinstance(value, str):
        return value
    raise ValidationError(f"'{name}' must be a string")


def opt_str(name: str, value: t.Any) -> str | None:
    if _guards.is_optional(value, str):
        return value
    raise ValidationError(f"'{name}' must be a string or null")


def opt_bool(name: str, value: t.Any) -> bool | None:
    if _guards.is_optional(value, bool):
        return value
    raise ValidationError(f"'{name}' must be a bool or null")


def str_list(name: str, value: t.Any) -> list[str]:
    if _guards.is_list_of(value, str):
        return value
    raise ValidationError(f"'{name}' must be a list of strings")


def opt_str_list(name: str, value: t.Any) -> list[str] | None:
    if _guards.is_optional_list_of(value, str):
        return value
    raise ValidationError(f"'{name}' must be a list of strings or null")


def opt_str_list_or_commasep(name: str, value: t.Any) -> list[str] | None:
    if _guards.is_optional_list_of(value, str):
        return value
    if isinstance(value, str):
        return value.split(",")
    raise ValidationError(
        f"'{name}' must be a list of strings or a comma-delimited string or null"
    )


def instance_or_dict(name: str, value: t.Any, cls: type[S]) -> S:
    if isinstance(value, cls):
        return value
    if isinstance(value, dict):
        return cls.from_dict(value)
    raise ValidationError(f"'{name}' must be a '{cls.__name__}' object or a dictionary")



from __future__ import annotations

import typing as t

from globus_sdk.scopes import ScopeBuilder


class SpecificFlowScopesClassStub(ScopeBuilder):
    """
    This stub object ensures that the type deductions for type checkers (e.g. mypy) on
    SpecificFlowClient.scopes are correct.

    Primarily, it should be possible to access the `scopes` attribute, the `user`
    scope, and the `resource_server`, but these usages should raise specific and
    informative runtime errors.

    Our types are therefore less accurate for class-var access, but more accurate for
    instance-var access.
    """

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__("<stub>")

    def __getattr__(self, name: str) -> t.Any:
        if name == "user":
            _raise_attr_error("scopes")
        elif name == "resource_server":
            _raise_attr_error("resource_server")
        return super().__getattr__(name)

    def __getattribute__(self, name: str) -> t.Any:
        if name == "resource_server":
            _raise_attr_error("resource_server")
        return object.__getattribute__(self, name)


def _raise_attr_error(name: str) -> t.NoReturn:
    raise AttributeError(
        f"It is not valid to attempt to access the {name} of the "
        "SpecificFlowClient class. "
        f"Instead, instantiate a SpecificFlowClient and access the {name} "
        "from that instance."
    )



from __future__ import annotations

import sys
import timeit
import warnings

from ._parser import parse_scope_graph

warnings.warn(
    "This is a test module for scope parsing. It is not intended for public use.",
    stacklevel=1,
)


def timeit_test() -> None:
    for size, num_iterations, style in (
        (10, 1000, "deep"),
        (100, 1000, "deep"),
        (1000, 1000, "deep"),
        (2000, 100, "deep"),
        (3000, 100, "deep"),
        (4000, 100, "deep"),
        (5000, 100, "deep"),
        (5000, 1000, "wide"),
        (10000, 1000, "wide"),
    ):
        if style == "deep":
            setup = f"""\
from globus_sdk.experimental.scope_parser import Scope
big_scope = ""
for i in range({size}):
    big_scope += f"foo{{i}}["
big_scope += "bar"
for _ in range({size}):
    big_scope += "]"
"""
        elif style == "wide":
            setup = f"""\
from globus_sdk.experimental.scope_parser import Scope
big_scope = ""
for i in range({size}):
    big_scope += f"foo{{i}} "
"""
        else:
            raise NotImplementedError

        timer = timeit.Timer("Scope.parse(big_scope)", setup=setup)

        raw_timings = timer.repeat(repeat=5, number=num_iterations)
        best, worst, average, variance = _stats(raw_timings)
        if style == "deep":
            print(f"{num_iterations} runs on a deep scope, depth={size}")
        elif style == "wide":
            print(f"{num_iterations} runs on a wide scope, width={size}")
        else:
            raise NotImplementedError
        print(f"  best={best} worst={worst} average={average} variance={variance}")
        print(f"  normalized best={best / num_iterations}")
        print()
    print("The most informative stat over these timings is the min timing (best).")
    print("Normed best is best/iterations.")
    print(
        "Max timing (worst) and dispersion (variance vis-a-vis average) indicate "
        "how consistent the results are, but are not a report of speed."
    )


def _stats(timing_data: list[float]) -> tuple[float, float, float, float]:
    best = min(timing_data)
    worst = max(timing_data)
    average = sum(timing_data) / len(timing_data)
    variance = sum((x - average) ** 2 for x in timing_data) / len(timing_data)
    return best, worst, average, variance


def parse_test(scope_string: str) -> None:
    parsed_graph = parse_scope_graph(scope_string)
    print(
        "top level scopes:",
        ", ".join([name for name, _optional in parsed_graph.top_level_scopes]),
    )
    print(parsed_graph)


def main() -> None:
    if len(sys.argv) < 2 or sys.argv[1] in ("-h", "--help"):
        print("This script supports two usage patterns:")
        print("    python -m globus_sdk.experimental.scope_parser SCOPE_STRING")
        print("    python -m globus_sdk.experimental.scope_parser --timeit")
        sys.exit(0)

    print()
    if sys.argv[1] == "--timeit":
        timeit_test()
    else:
        parse_test(sys.argv[1])


main()



from __future__ import annotations

import typing as t

from globus_sdk import utils
from globus_sdk._types import UUIDLike


class UserCredentialDocument(utils.PayloadWrapper):
    """
    Convenience class for constructing a UserCredential document
    to use as the `data` parameter to `create_user_credential` and
    `update_user_credential`

    :param DATA_TYPE: Versioned document type.
    :param identity_id: UUID of the Globus identity this credential will
        provide access for
    :param connector_id: UUID of the connector this credential is for
    :param username: Username of the local account this credential will provide
        access to, format is connector specific
    :param display_name: Display name for this credential
    :param storage_gateway_id: UUID of the storage gateway this credential is for
    :param policies: Connector specific policies for this credential
    :param additional_fields: Additional data for inclusion in the document
    """

    def __init__(
        self,
        DATA_TYPE: str = "user_credential#1.0.0",
        identity_id: UUIDLike | None = None,
        connector_id: UUIDLike | None = None,
        username: str | None = None,
        display_name: str | None = None,
        storage_gateway_id: UUIDLike | None = None,
        policies: dict[str, t.Any] | None = None,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self._set_optstrs(
            DATA_TYPE=DATA_TYPE,
            identity_id=identity_id,
            connector_id=connector_id,
            username=username,
            display_name=display_name,
            storage_gateway_id=storage_gateway_id,
        )
        self._set_value("policies", policies)

        if additional_fields is not None:
            self.update(additional_fields)



"""
Definition and loading of standard environment variables, plus a wrappers for loading
and parsing values.

This does not include service URL env vars (see environments.py for loading of those)
"""

from __future__ import annotations

import logging
import os
import pathlib
import typing as t

log = logging.getLogger(__name__)
T = t.TypeVar("T")


ENVNAME_VAR = "GLOBUS_SDK_ENVIRONMENT"
HTTP_TIMEOUT_VAR = "GLOBUS_SDK_HTTP_TIMEOUT"
SSL_VERIFY_VAR = "GLOBUS_SDK_VERIFY_SSL"


@t.overload
def _load_var(
    varname: str,
    default: t.Any,
    explicit_value: t.Any | None,
    convert: t.Callable[[t.Any, t.Any], T],
) -> T: ...


@t.overload
def _load_var(
    varname: str,
    default: str,
    explicit_value: str | None,
) -> str: ...


def _load_var(
    varname: str,
    default: t.Any,
    explicit_value: t.Any | None = None,
    convert: t.Callable[[t.Any, t.Any], T] | None = None,
) -> t.Any:
    # use the explicit value if given and non-None, otherwise, do an env lookup
    value = (
        explicit_value if explicit_value is not None else os.getenv(varname, default)
    )
    if convert:
        value = convert(value, default)
    # only info log on non-default *values*
    # meaning that if we define the default as 'foo' and someone explicitly sets 'foo',
    # no info log gets emitted
    if value != default:
        log.info(f"on lookup, non-default setting: {varname}={value}")
    else:
        log.debug(f"on lookup, default setting: {varname}={value}")
    return value


def _ssl_verify_cast(
    value: t.Any, default: t.Any  # pylint: disable=unused-argument
) -> bool | str:
    if isinstance(value, bool):
        return value
    if not isinstance(value, (str, pathlib.Path)):
        msg = f"Value {value} of type {type(value)} cannot be used for SSL verification"
        raise ValueError(msg)
    if isinstance(value, str):
        if value.lower() in {"y", "yes", "t", "true", "on", "1"}:
            return True
        if value.lower() in {"n", "no", "f", "false", "off", "0"}:
            return False
        if os.path.isfile(value):
            return value
    if isinstance(value, pathlib.Path) and value.is_file():
        return str(value.absolute())
    raise ValueError(
        "SSL verification value must be a valid boolean value "
        f"or a path to a file that exists (got {value})"
    )


def _optfloat_cast(value: t.Any, default: t.Any) -> float | None:
    try:
        return float(value)
    except ValueError:
        pass
    if value == "":
        return t.cast(float, default)
    log.error(f'Value "{value}" can\'t cast to optfloat')
    raise ValueError(f"Invalid config float: {value}")


def get_environment_name(inputenv: str | None = None) -> str:
    return _load_var(ENVNAME_VAR, "production", explicit_value=inputenv)


def get_ssl_verify(value: bool | str | pathlib.Path | None = None) -> bool | str:
    return _load_var(
        SSL_VERIFY_VAR, default=True, explicit_value=value, convert=_ssl_verify_cast
    )


def get_http_timeout(value: float | None = None) -> float | None:
    ret = _load_var(
        HTTP_TIMEOUT_VAR, 60.0, explicit_value=value, convert=_optfloat_cast
    )
    if ret == -1.0:
        return None
    return ret



#!/usr/bin/env python
from __future__ import annotations

import itertools
import pathlib
import textwrap
import typing as t

HERE = pathlib.Path(__file__).parent

FIXED_PREAMBLE = f"""\
# isort:skip_file
# fmt:off
#
# this __init__.py file is generated by {pathlib.Path(__file__).name}
# do not edit it directly or testing will fail
import importlib
import logging
import sys
import typing as t

from .version import __version__


def _force_eager_imports() -> None:
    current_module = sys.modules[__name__]

    for attribute_set in _LAZY_IMPORT_TABLE.values():
        for attr in attribute_set:
            getattr(current_module, attr)
"""

FIXED_EPILOG = """
# configure logging for a library, per python best practices:
# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger("globus_sdk").addHandler(logging.NullHandler())
"""

FIXED_MODULE_METHODS = """
def __dir__() -> t.List[str]:
    # dir(globus_sdk) should include everything exported in __all__
    # as well as some explicitly selected attributes from the default dir() output
    # on a module
    #
    # see also:
    # https://discuss.python.org/t/how-to-properly-extend-standard-dir-search-with-module-level-dir/4202
    return list(__all__) + [
        # __all__ itself can be inspected
        "__all__",
        # useful to figure out where a package is installed
        "__file__",
        "__path__",
    ]


def __getattr__(name: str) -> t.Any:
    for modname, items in _LAZY_IMPORT_TABLE.items():
        if name in items:
            mod = importlib.import_module("." + modname, __name__)
            value = getattr(mod, name)
            setattr(sys.modules[__name__], name, value)
            return value

    raise AttributeError(f"module {__name__} has no attribute {name}")
"""


_LAZY_IMPORT_TABLE: list[tuple[str, tuple[str, ...]]] = [
    (
        "authorizers",
        (
            "AccessTokenAuthorizer",
            "BasicAuthorizer",
            "ClientCredentialsAuthorizer",
            "NullAuthorizer",
            "RefreshTokenAuthorizer",
        ),
    ),
    ("client", ("BaseClient",)),
    (
        "exc",
        (
            "GlobusAPIError",
            "ErrorSubdocument",
            "GlobusConnectionError",
            "GlobusConnectionTimeoutError",
            "GlobusError",
            "GlobusSDKUsageError",
            "GlobusTimeoutError",
            "NetworkError",
            "RemovedInV4Warning",
        ),
    ),
    (
        "local_endpoint",
        (
            "GlobusConnectPersonalOwnerInfo",
            "LocalGlobusConnectPersonal",
            "LocalGlobusConnectServer",
        ),
    ),
    (
        "response",
        (
            "GlobusHTTPResponse",
            "IterableResponse",
            "ArrayResponse",
        ),
    ),
    (
        "services.auth",
        (
            # client classes
            "AuthClient",
            "AuthLoginClient",
            "NativeAppAuthClient",
            "ConfidentialAppAuthClient",
            # errors
            "AuthAPIError",
            # high-level helpers
            "IdentityMap",
            # responses
            "GetIdentitiesResponse",
            "OAuthDependentTokenResponse",
            "OAuthTokenResponse",
            # API data helpers
            "DependentScopeSpec",
        ),
    ),
    (
        "services.gcs",
        (
            "CollectionDocument",
            "GCSAPIError",
            "GCSClient",
            "GCSRoleDocument",
            "EndpointDocument",
            "GuestCollectionDocument",
            "MappedCollectionDocument",
            "CollectionPolicies",
            "POSIXCollectionPolicies",
            "POSIXStagingCollectionPolicies",
            "GoogleCloudStorageCollectionPolicies",
            "StorageGatewayDocument",
            "StorageGatewayPolicies",
            "POSIXStoragePolicies",
            "POSIXStagingStoragePolicies",
            "BlackPearlStoragePolicies",
            "BoxStoragePolicies",
            "CephStoragePolicies",
            "GoogleDriveStoragePolicies",
            "GoogleCloudStoragePolicies",
            "OneDriveStoragePolicies",
            "AzureBlobStoragePolicies",
            "S3StoragePolicies",
            "ActiveScaleStoragePolicies",
            "IrodsStoragePolicies",
            "HPSSStoragePolicies",
            "UserCredentialDocument",
            "IterableGCSResponse",
            "UnpackingGCSResponse",
        ),
    ),
    (
        "services.flows",
        (
            "FlowsClient",
            "FlowsAPIError",
            "IterableFlowsResponse",
            "SpecificFlowClient",
        ),
    ),
    (
        "services.groups",
        (
            "BatchMembershipActions",
            "GroupMemberVisibility",
            "GroupPolicies",
            "GroupRequiredSignupFields",
            "GroupRole",
            "GroupsAPIError",
            "GroupsClient",
            "GroupsManager",
            "GroupVisibility",
        ),
    ),
    (
        "services.search",
        (
            "SearchAPIError",
            "SearchClient",
            "SearchQuery",
            "SearchScrollQuery",
        ),
    ),
    (
        "services.timer",
        (
            "TimerAPIError",
            "TimerClient",
            "TransferTimer",
            "TimerJob",
            "OnceTimerSchedule",
            "RecurringTimerSchedule",
        ),
    ),
    (
        "services.transfer",
        (
            "ActivationRequirementsResponse",
            "DeleteData",
            "IterableTransferResponse",
            "TransferAPIError",
            "TransferClient",
            "TransferData",
        ),
    ),
    (
        "utils",
        (
            "MISSING",
            "MissingType",
        ),
    ),
]


def _generate_imports() -> t.Iterator[str]:
    for modname, items in _LAZY_IMPORT_TABLE:
        for item in items:
            yield textwrap.indent(f"from .{modname} import {item}", "    ")


def _generate_lazy_import_table() -> t.Iterator[str]:
    yield "_LAZY_IMPORT_TABLE = {"
    for modname, items in _LAZY_IMPORT_TABLE:
        yield textwrap.indent(f'"{modname}": {{', " " * 4)
        for item in items:
            yield textwrap.indent(f'"{item}",', " " * 8)
        yield textwrap.indent("},", " " * 4)
    yield "}"


def _generate_all_tuple() -> t.Iterator[str]:
    yield "__all__ = ("
    yield '    "__version__",'
    yield '    "_force_eager_imports",'
    yield from (
        f'    "{item}",'
        for item in sorted(itertools.chain(*[items for _, items in _LAZY_IMPORT_TABLE]))
    )
    yield ")"


def _init_pieces() -> t.Iterator[str]:
    yield FIXED_PREAMBLE
    yield ""
    yield from _generate_lazy_import_table()
    yield ""
    yield "if t.TYPE_CHECKING:"
    yield from _generate_imports()
    yield ""
    yield FIXED_MODULE_METHODS
    yield ""
    yield from _generate_all_tuple()
    yield ""
    yield FIXED_EPILOG


def _generate_init() -> str:
    return "\n".join(_init_pieces())


def main() -> None:
    with open(HERE / "__init__.py", "w", encoding="utf-8") as fp:
        fp.write(_generate_init())


if __name__ == "__main__":
    main()



from __future__ import annotations

# the name "datetime" is used in this module, so use an alternative name
# in order to avoid name shadowing
import datetime as dt
import logging
import typing as t

from globus_sdk.config import get_service_url
from globus_sdk.exc import warn_deprecated
from globus_sdk.services.transfer import TransferData
from globus_sdk.utils import MISSING, MissingType, PayloadWrapper, slash_join

log = logging.getLogger(__name__)


class TransferTimer(PayloadWrapper):
    """
    A helper for defining a payload for Transfer Timer creation.
    Use this along with :meth:`create_timer <globus_sdk.TimerClient.create_timer>` to
    create a timer.

    .. note::
        ``TimerClient`` has two methods for creating timers, ``create_timer`` and
        ``create_job``.
        ``create_job`` uses a different API -- only ``create_timer`` will work with
        this helper class.

        Users are strongly recommended to use ``create_timer`` and this helper for
        timer creation.

    :param name: A name to identify this timer
    :param schedule: The schedule on which the timer runs
    :param body: A transfer payload for the timer to use. If it includes
        ``submission_id`` or ``skip_activation_check``, these parameters will be
        removed, as they are not supported in timers.

    The ``schedule`` field determines when the timer will run.
    Timers may be "run once" or "recurring", and "recurring" timers may specify an end
    date or a number of executions after which the timer will stop. A ``schedule`` is
    specified as a dict, but the SDK provides two useful helpers for constructing these
    data.

    **Example Schedules**

    .. tab-set::

        .. tab-item:: Run Once, Right Now

            .. code-block:: python

                schedule = OnceTimerSchedule()

        .. tab-item:: Run Once, At a Specific Time

            .. code-block:: python

                schedule = OnceTimerSchedule(datetime="2023-09-22T00:00:00Z")

        .. tab-item:: Run Every 5 Minutes, Until a Specific Time

            .. code-block:: python

                schedule = RecurringTimerSchedule(
                    interval_seconds=300,
                    end={"condition": "time", "datetime": "2023-10-01T00:00:00Z"},
                )

        .. tab-item:: Run Every 30 Minutes, 10 Times

            .. code-block:: python

                schedule = RecurringTimerSchedule(
                    interval_seconds=1800,
                    end={"condition": "iterations", "iterations": 10},
                )

        .. tab-item:: Run Every 10 Minutes, Indefinitely

            .. code-block:: python

                schedule = RecurringTimerSchedule(interval_seconds=600)

    Using these schedules, you can create a timer from a ``TransferData`` object:

    .. code-block:: pycon

        >>> from globus_sdk import TransferData, TransferTimer
        >>> schedule = ...
        >>> transfer_data = TransferData(...)
        >>> timer = TransferTimer(
        ...     name="my timer",
        ...     schedule=schedule,
        ...     body=transfer_data,
        ... )

    Submit the timer to the Timers service with
    :meth:`create_timer <globus_sdk.TimerClient.create_timer>`.
    """

    def __init__(
        self,
        *,
        name: str | MissingType = MISSING,
        schedule: dict[str, t.Any] | RecurringTimerSchedule | OnceTimerSchedule,
        body: dict[str, t.Any] | TransferData,
    ) -> None:
        super().__init__()
        self["timer_type"] = "transfer"
        self["name"] = name
        self["schedule"] = schedule
        self["body"] = self._preprocess_body(body)

    def _preprocess_body(
        self, body: dict[str, t.Any] | TransferData
    ) -> dict[str, t.Any]:
        # shallow-copy for dicts, convert any TransferData to a dict
        new_body = dict(body)
        # remove the skip_activation_check and submission_id parameters unconditionally
        # (not supported in timers, but often present in TransferData)
        new_body.pop("submission_id", None)
        new_body.pop("skip_activation_check", None)
        return new_body


class RecurringTimerSchedule(PayloadWrapper):
    """
    A helper used as part of a *timer* to define when the *timer* will run.

    A ``RecurringTimerSchedule`` is used to describe a *timer* which runs repeatedly
    until some end condition is reached.

    :param interval_seconds: The number of seconds between each run of the timer.
    :param start: The time at which to start the timer, either as an ISO 8601 string
        with timezone information, or as a ``datetime.datetime`` object.
    :param end: The end condition for the timer, as a dict. This either expresses a
        number of iterations for the timer or an end date.

    Example ``end`` conditions:

    .. code-block:: python

        # run 10 times
        end = {"condition": "iterations", "iterations": 10}

        # run until a specific date
        end = {"condition": "time", "datetime": "2023-10-01T00:00:00Z"}

    If the end condition is ``time``, then the ``datetime`` value can be expressed as a
    python ``datetime`` type as well, e.g.

    .. code-block:: python

        # end in 10 days
        end = {
            "condition": "time",
            "datetime": datetime.datetime.now() + datetime.timedelta(days=10),
        }
    """

    def __init__(
        self,
        interval_seconds: int,
        start: str | dt.datetime | MissingType = MISSING,
        end: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["type"] = "recurring"
        self["interval_seconds"] = interval_seconds
        self["start"] = _format_date(start)
        self["end"] = end

        # if a datetime is given for part of the end condition, format it (and
        # shallow-copy the end condition)
        # primarily, this handles
        #    end={"condition": "time", "datetime": <some-datetime>}
        if isinstance(end, dict):
            self["end"] = {
                k: (_format_date(v) if isinstance(v, dt.datetime) else v)
                for k, v in end.items()
            }


class OnceTimerSchedule(PayloadWrapper):
    """
    A helper used as part of a *timer* to define when the *timer* will run.

    A ``OnceTimerSchedule`` is used to describe a *timer* which runs exactly once.
    It may be scheduled for a time in the future.

    :param datetime: The time at which to run the timer, either as an ISO 8601
        string with timezone information, or as a ``datetime.datetime`` object.
    """

    def __init__(
        self,
        datetime: str | dt.datetime | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["type"] = "once"
        self["datetime"] = _format_date(datetime)


class TimerJob(PayloadWrapper):
    r"""
    .. warning::

        This method of specifying and creating Timers for data transfer is now
        deprecated. Users should use ``TimerData`` instead.

        ``TimerJob`` is still supported for non-transfer use-cases.

    Helper for creating a timer in the Timer service. Used as the ``data``
    argument in :meth:`create_job <globus_sdk.TimerClient.create_job>`.

    The ``callback_url`` parameter should always be the URL used to run an
    action provider.

    :param callback_url: URL for the action which the Timer job will use.
    :param callback_body: JSON data which Timer will send to the Action Provider on
        each invocation
    :param start: The datetime at which to start the Timer job.
    :param interval: The interval at which the Timer job should recur. Interpreted as
        seconds if specified as an integer. If ``stop_after_n == 1``, i.e. the job is
        set to run only a single time, then interval *must* be None.
    :param name: A (not necessarily unique) name to identify this job in Timer
    :param stop_after: A date after which the Timer job will stop running
    :param stop_after_n: A number of executions after which the Timer job will stop
    :param scope: Timer defaults to the Transfer 'all' scope. Use this parameter to
        change the scope used by Timer when calling the Transfer Action Provider.

    .. automethodlist:: globus_sdk.TimerJob
    """

    def __init__(
        self,
        callback_url: str,
        callback_body: dict[str, t.Any],
        start: dt.datetime | str,
        interval: dt.timedelta | int | None,
        *,
        name: str | None = None,
        stop_after: dt.datetime | None = None,
        stop_after_n: int | None = None,
        scope: str | None = None,
    ) -> None:
        super().__init__()
        self["callback_url"] = callback_url
        self["callback_body"] = callback_body
        if isinstance(start, dt.datetime):
            self["start"] = start.isoformat()
        else:
            self["start"] = start
        if isinstance(interval, dt.timedelta):
            self["interval"] = int(interval.total_seconds())
        else:
            self["interval"] = interval
        if name is not None:
            self["name"] = name
        if stop_after is not None:
            self["stop_after"] = stop_after.isoformat()
        if stop_after_n is not None:
            self["stop_after_n"] = stop_after_n
        if scope is not None:
            self["scope"] = scope

    @classmethod
    def from_transfer_data(
        cls,
        transfer_data: TransferData | dict[str, t.Any],
        start: dt.datetime | str,
        interval: dt.timedelta | int | None,
        *,
        name: str | None = None,
        stop_after: dt.datetime | None = None,
        stop_after_n: int | None = None,
        scope: str | None = None,
        environment: str | None = None,
    ) -> TimerJob:
        r"""
        Specify data to create a Timer job using the parameters for a transfer. Timer
        will use those parameters to run the defined transfer operation, recurring at
        the given interval.

        :param transfer_data: A :class:`TransferData <globus_sdk.TransferData>` object.
            Construct this object exactly as you would normally; Timer will use this to
            run the recurring transfer.
        :param start: The datetime at which to start the Timer job.
        :param interval: The interval at which the Timer job should recur. Interpreted
            as seconds if specified as an integer. If ``stop_after_n == 1``, i.e. the
            job is set to run only a single time, then interval *must* be None.
        :param name: A (not necessarily unique) name to identify this job in Timer
        :param stop_after: A date after which the Timer job will stop running
        :param stop_after_n: A number of executions after which the Timer job will stop
        :param scope: Timer defaults to the Transfer 'all' scope. Use this parameter to
            change the scope used by Timer when calling the Transfer Action Provider.
        :param environment: For internal use: because this method needs to generate a
            URL for the Transfer Action Provider, this argument can control which
            environment the Timer job is sent to.
        """
        warn_deprecated(
            "TimerJob.from_transfer_data(X, ...) is deprecated. "
            "Prefer TransferTimer(body=X, ...) instead."
        )

        transfer_action_url = slash_join(
            get_service_url("actions", environment=environment), "transfer/transfer/run"
        )
        log.info(
            "Creating TimerJob from TransferData, action_url=%s", transfer_action_url
        )
        for key in ("submission_id", "skip_activation_check"):
            if key in transfer_data:
                raise ValueError(
                    f"cannot create TimerJob from TransferData which has {key} set"
                )
        # dict will either convert a `TransferData` object or leave us with a dict here
        callback_body = {"body": dict(transfer_data)}
        return cls(
            transfer_action_url,
            callback_body,
            start,
            interval,
            name=name,
            stop_after=stop_after,
            stop_after_n=stop_after_n,
            scope=scope,
        )


def _format_date(date: str | dt.datetime | MissingType) -> str | MissingType:
    if isinstance(date, dt.datetime):
        if date.tzinfo is None:
            date = date.astimezone(dt.timezone.utc)
        return date.isoformat(timespec="seconds")
    else:
        return date
